@@ -11,7 +11,7 @@ use core::{
     hash::{BuildHasher, Hash, Hasher},
 };
 use hashbrown::{raw::RawTable, HashSet};
-use std::{collections::hash_map::RandomState, borrow::Borrow};
+use std::{borrow::Borrow, collections::hash_map::RandomState};
 
 mod language;
 pub use language::*;
@@ -106,7 +106,7 @@ where
     I: Copy + Eq,
 {
     /// Create a new, empty congruence closure
-    /// 
+    ///
     /// # Example
     /// ```rust
     /// # use congruence::CongruenceClosure;
@@ -123,7 +123,7 @@ where
     }
 
     /// Create a new, empty congruence closure with the given node and pair capacities
-    /// 
+    ///
     /// # Example
     /// ```rust
     /// # use congruence::CongruenceClosure;
@@ -173,7 +173,7 @@ where
     }
 
     /// Whether this congruence closure is empty, i.e. contains no *congruence* relations
-    /// 
+    ///
     /// # Example
     /// ```rust
     /// # use congruence::{CongruenceClosure, DisjointSetForest, CongruenceState};
@@ -181,19 +181,19 @@ where
     /// let mut dsf = DisjointSetForest::new();
     /// let mut cs = CongruenceState::new();
     /// assert!(cc.is_empty());
-    /// 
+    ///
     /// // This operation only adds something to the disjoint set forest, leaving the congruence
     /// // closure itself empty
     /// cc.merge(3, 4, &mut dsf, &mut cs);
     /// assert!(cc.is_empty());
     /// assert!(!dsf.is_empty());
-    /// 
+    ///
     /// // We now actually add something to the congruence closure by equating an *expression*, here
     /// // simply a pair of `usize` indices, to a term
     /// cc.equation((1, 2), 3, &mut dsf, &mut cs);
     /// assert!(!cc.is_empty());
     /// assert!(!dsf.is_empty());
-    /// 
+    ///
     /// // On the other hand, equating an expression to a term without introducing any new term-level
     /// // equalities only adds things to the congruence closure, leaving the union-find structure
     /// // unchanged:
@@ -209,7 +209,7 @@ where
     }
 
     /// Clear this congruence closure, maintaining it's capacity but otherwise resetting it to empty
-    /// 
+    ///
     /// # Example
     /// ```rust
     /// # use congruence::{CongruenceClosure, DisjointSetForest, CongruenceState};
@@ -229,7 +229,7 @@ where
     }
 
     /// Register an equation of the form `expr = result`, where `expr` is an expression in this congruence closure's language
-    /// 
+    ///
     /// # Example
     /// ```rust
     /// # use congruence::{CongruenceClosure, DisjointSetForest, CongruenceState, UnionFind};
@@ -242,7 +242,7 @@ where
     /// assert!(dsf.node_eq(0, 1));
     /// assert!(dsf.node_eq(2, 3));
     /// assert!(!dsf.node_eq(4, 5));
-    /// 
+    ///
     /// // If we set (0, 2) ~ 4 and (1, 3) ~ 5, we can then deduce 4 ~ 5
     /// cc.equation((0, 2), 4, &mut dsf, &mut cs);
     /// assert!(dsf.node_eq(0, 1));
@@ -263,7 +263,11 @@ where
         L: Language<U, I>,
         U: UnionFind<I>,
     {
-        let hash = expr.hash_one_mod_mut(&self.hasher, &mut |uf, ix| uf.find_mut(ix), union_find);
+        let hash = expr.hash_one_mod_mut(
+            &self.hasher,
+            &mut |hs, uf, ix| uf.find_mut(ix).hash(hs),
+            union_find,
+        );
         if let Some(&(_, b)) = self.lookup.get(hash, |(k, _r)| {
             expr.eq_mod_mut(k, &mut |uf, i, j| uf.node_eq_mut(i, j), union_find)
         }) {
@@ -275,13 +279,17 @@ where
             );
             debug_assert_eq!(r, Ok(()));
             self.lookup.insert(hash, (expr, result), |(k, _r)| {
-                k.hash_one_mod(&self.hasher, &mut |uf, ix| uf.find(ix), union_find)
+                k.hash_one_mod(
+                    &self.hasher,
+                    &mut |hs, uf, ix| uf.find(ix).hash(hs),
+                    union_find,
+                )
             });
         }
     }
 
     /// Get a representative for the congruence class of `expr` in this context, if one exists.
-    /// 
+    ///
     /// # Example
     /// ```rust
     /// # use congruence::{CongruenceClosure, DisjointSetForest, CongruenceState};
@@ -307,14 +315,18 @@ where
         L: Borrow<B>,
         U: UnionFind<I>,
     {
-        let hash = expr.hash_one_mod(&self.hasher, &mut |uf, ix| uf.find(ix), union_find);
+        let hash = expr.hash_one_mod(
+            &self.hasher,
+            &mut |hs, uf, ix| uf.find(ix).hash(hs),
+            union_find,
+        );
         self.lookup.get(hash, |(k, _r)| {
             expr.eq_mod(k.borrow(), &mut |uf, i, j| uf.node_eq(i, j), union_find)
         })
     }
 
     /// Get a representative for the congruence class of `expr` in this context, if one exists.
-    /// 
+    ///
     /// Always returns the same result as [`Self::lookup`], but may optimize the union-find data structure provided in the process.
     pub fn lookup_mut<B, U>(&self, expr: &B, union_find: &mut U) -> Option<&(L, I)>
     where
@@ -322,29 +334,33 @@ where
         L: Borrow<B>,
         U: UnionFind<I>,
     {
-        let hash = expr.hash_one_mod_mut(&self.hasher, &mut |uf, ix| uf.find_mut(ix), union_find);
+        let hash = expr.hash_one_mod_mut(
+            &self.hasher,
+            &mut |hs, uf, ix| uf.find_mut(ix).hash(hs),
+            union_find,
+        );
         self.lookup.get(hash, |(k, _r)| {
             expr.eq_mod_mut(k.borrow(), &mut |uf, i, j| uf.node_eq_mut(i, j), union_find)
         })
     }
 
     /// Check whether an expressions in the equation language is congruent to a given index
-    /// 
+    ///
     /// # Example
     /// ```rust
     /// # use congruence::{CongruenceClosure, DisjointSetForest, CongruenceState};
     /// let mut cc = CongruenceClosure::<(usize, usize)>::new();
     /// let mut dsf = DisjointSetForest::new();
     /// let mut cs = CongruenceState::new();
-    /// 
+    ///
     /// assert!(!cc.expr_cong(&(0, 1), 4, &dsf));
     /// assert!(!cc.expr_cong(&(2, 3), 4, &dsf));
-    /// 
+    ///
     /// cc.merge(0, 2, &mut dsf, &mut cs);
     /// cc.merge(1, 3, &mut dsf, &mut cs);
     /// assert!(!cc.expr_cong(&(0, 1), 4, &dsf));
     /// assert!(!cc.expr_cong(&(2, 3), 4, &dsf));
-    /// 
+    ///
     /// // If we set `(0, 1) ~ 4`, we can deduce `(2, 3) ~ 4`, even though we've never inserted the pair `(2, 3)`
     /// cc.equation((0, 1), 4, &mut dsf, &mut cs);
     /// assert!(cc.expr_cong(&(0, 1), 4, &dsf));
@@ -357,13 +373,13 @@ where
         U: UnionFind<I>,
     {
         if let Some((_, repr)) = self.lookup(expr, union_find) {
-                return union_find.node_eq(*repr, ix);
+            return union_find.node_eq(*repr, ix);
         }
         false
     }
 
     /// Check whether an expressions in the equation language is congruent to a given index
-    /// 
+    ///
     /// Always returns the same result as [`Self::expr_cong`], but may optimize the union-find data structure provided in the process.
     pub fn expr_cong_mut<B, U>(&self, expr: &B, ix: I, union_find: &mut U) -> bool
     where
@@ -372,13 +388,13 @@ where
         U: UnionFind<I>,
     {
         if let Some((_, repr)) = self.lookup_mut(expr, union_find) {
-                return union_find.node_eq_mut(*repr, ix);
+            return union_find.node_eq_mut(*repr, ix);
         }
         false
     }
 
     /// Merge the equivalence classes of two nodes
-    /// 
+    ///
     /// # Example    
     /// ```rust
     /// # use congruence::{CongruenceClosure, DisjointSetForest, CongruenceState, UnionFind};
@@ -386,32 +402,32 @@ where
     /// let mut dsf = DisjointSetForest::new();
     /// let mut raw_dsf = DisjointSetForest::new();
     /// let mut cs = CongruenceState::new();
-    /// 
+    ///
     /// // Calls to `cc.merge` without any equations added to the congruence context are equivalent to calls to `dsf.union`
     /// cc.merge(0, 1, &mut dsf, &mut cs);
     /// raw_dsf.union(0, 1);
     /// assert_eq!(dsf, raw_dsf);
-    /// 
+    ///
     /// cc.merge(2, 3, &mut dsf, &mut cs);
     /// raw_dsf.union(2, 3);
     /// assert_eq!(dsf, raw_dsf);
-    /// 
+    ///
     /// // Similarly, with only one equation in the context, they do not change the union-find ADT,
     /// // though the equation itself may be modified as the union-find is updated:
     /// assert_eq!(cc.lookup(&(4, 5), &dsf), None);
     /// cc.equation((4, 5), 6, &mut dsf, &mut cs);
     /// assert_eq!(dsf, raw_dsf);
     /// assert_eq!(cc.lookup(&(4, 5), &dsf), Some(&((4, 5), 6)));
-    /// 
+    ///
     /// cc.merge(0, 4, &mut dsf, &mut cs);
     /// raw_dsf.union(0, 4);
     /// assert_eq!(dsf, raw_dsf);
-    /// 
+    ///
     /// // Once we have a set of equations which overlap, however, calling `merge` can trigger multiple `unions` in the DSF,
     /// // implementing congruence closure:
     /// cc.equation((4, 7), 8, &mut dsf, &mut cs);
     /// assert_eq!(dsf, raw_dsf);
-    /// 
+    ///
     /// cc.merge(7, 5, &mut dsf, &mut cs);
     /// raw_dsf.union(7, 5);
     /// assert!(raw_dsf.refines(&dsf));
@@ -448,20 +464,16 @@ where
                 for old_key in list.iter().flatten() {
                     let curr_hash = old_key.hash_one_mod_mut(
                         &self.hasher,
-                        &mut |uf, ix| uf.find_mut(ix),
+                        &mut |hs, uf, ix| uf.find_mut(ix).hash(hs),
                         union_find,
                     );
                     let mut has_new_repr = false;
                     let new_hash = old_key.hash_one_mod_mut(
                         &self.hasher,
-                        &mut |uf, ix| {
+                        &mut |hs, uf, ix| {
                             let found = uf.find_mut(ix);
                             has_new_repr |= found == new_repr;
-                            if found == old_repr {
-                                new_repr
-                            } else {
-                                found
-                            }
+                            if found == old_repr { new_repr } else { found }.hash(hs)
                         },
                         union_find,
                     );
@@ -492,13 +504,10 @@ where
                                 .insert(new_hash, (old_key.clone(), c), |(k, _r)| {
                                     k.hash_one_mod(
                                         &self.hasher,
-                                        &mut |uf, i| {
+                                        &mut |hs, uf, i| {
                                             let found = uf.find(i);
-                                            if found == old_repr {
-                                                new_repr
-                                            } else {
-                                                found
-                                            }
+                                            if found == old_repr { new_repr } else { found }
+                                                .hash(hs)
                                         },
                                         union_find,
                                     )
@@ -538,8 +547,11 @@ where
                     return false;
                 }
                 for user in users {
-                    let hash =
-                        user.hash_one_mod_mut(&self.hasher, &mut |uf, ix| uf.find(ix), union_find);
+                    let hash = user.hash_one_mod_mut(
+                        &self.hasher,
+                        &mut |hs, uf, ix| uf.find(ix).hash(hs),
+                        union_find,
+                    );
                     if let Some(lookup) = self.lookup.get(hash, |(k, _r)| {
                         k.eq_mod_mut(user, &mut |uf, i, j| uf.node_eq_mut(i, j), union_find)
                     }) {
